/*
    Sliced vbi demonstration utility
    Copyright (C) 2004  Hans Verkuil  <hverkuil@xs4all.nl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <unistd.h>
#include <features.h> /* Uses _GNU_SOURCE to define getsubopt in stdlib.h */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <inttypes.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <math.h>
#include <linux/types.h>

#include "videodev2.h"

#include "ivtv.h"

#define printable(c) ((((c) & 0x7F) < 0x20 || ((c) & 0x7F) > 0x7E) ? '.' : ((c) & 0x7F))

static const char *formats[] = {
	"Full format 4:3, 576 lines",
	"Letterbox 14:9 centre, 504 lines",
	"Letterbox 14:9 top, 504 lines",
	"Letterbox 16:9 centre, 430 lines",
	"Letterbox 16:9 top, 430 lines",
	"Letterbox > 16:9 centre",
	"Full format 14:9 centre, 576 lines",
	"Anamorphic 16:9, 576 lines"
};
static const char *subtitles[] = {
	"none",
	"in active image area",
	"out of active image area",
	"?"
};

void decode_wss(struct ivtv_sliced_data *s)
{
	unsigned char parity;
	int wss;

	wss = s->data[0] | (s->data[1] << 8);
	
	parity = wss & 15;
	parity ^= parity >> 2;
	parity ^= parity >> 1;

	if (!(parity & 1)) return;

	printf("WSS: %s; %s mode; %s color coding;\n"
	       "      %s helper; reserved b7=%d; %s\n"
	       "      open subtitles: %s; %scopyright %s; copying %s\n",
	       formats[wss & 7],
	       (wss & 0x10) ? "film" : "camera",
	       (wss & 0x20) ? "MA/CP" : "standard",
	       (wss & 0x40) ? "modulated" : "no",
	       !!(wss & 0x80),
	       (wss & 0x0100) ? "have TTX subtitles; " : "",
	       subtitles[(wss >> 9) & 3],
	       (wss & 0x0800) ? "surround sound; " : "",
	       (wss & 0x1000) ? "asserted" : "unknown",
	       (wss & 0x2000) ? "restricted" : "not restricted");
}

static int odd_parity(uint8_t c)
{
	c ^= (c >> 4);
	c ^= (c >> 2);
	c ^= (c >> 1);

	return c & 1;
}

static void decode_xds(struct ivtv_sliced_data *s)
{
	char c;

	printf("XDS: %02x %02x: ", s->data[0], s->data[1]);
	c = odd_parity(s->data[0]) ? s->data[0] & 0x7F : '?';
	c = printable(c);
	putchar(c);
	c = odd_parity(s->data[1]) ? s->data[1] & 0x7F : '?';
	c = printable(c);
	putchar(c);
	putchar('\n');
}

#define CC_SIZE 70

//ccdecode
char    *ratings[] = {"(NOT RATED)","TV-Y","TV-Y7","TV-G","TV-PG","TV-14","TV-MA","(NOT RATED)"};
int     rowdata[] = {11,-1,1,2,3,4,12,13,14,15,5,6,7,8,9,10};
char    *modes[]={"current","future","channel","miscellanious","public service","reserved","invalid","invalid","invalid","invalid"};
char	*specialchar[] = {"®","°","½","¿","(TM)","¢","£","o/~ ","à"," ","è","â","ê","î","ô","û"};
int     lastcode;
int     ccmode=1;               //cc1 or cc2
char    ccbuf[3][256];          //cc is 32 columns per row, this allows for extra characters
int     keywords=0;
char    *keyword[32];


int CCdecode(int b1, int b2, int len)
{
        int x,y, row;

        if ((b1&0x10) && (b2>0x1F)) 
	//codes are always transmitted twice 
	//(apparently not, ignore the second occurance)
        {
                ccmode=((b1>>3)&1)+1;
                len = strlen(ccbuf[ccmode]);

                if (b2 & 0x40)  //preamble address code (row & indent)
                {
                        row=rowdata[((b1<<1)&14)|((b2>>5)&1)];
                        if (len!=0)
                                ccbuf[ccmode][len++]='\n';

                        if (b2&0x10) //row contains indent flag
                                for (x=0;x<(b2&0x0F)<<1;x++)
                                        ccbuf[ccmode][len++]=' ';
                }
                else
                {
                        switch (b1 & 0x07)
                        {
                                case 0x00:      //attribute
                                        printf("<ATTRIBUTE %d %d>\n",b1,b2);
                                        fflush(stdout);
                                        break;
                                case 0x01:      //midrow or char
                                        switch (b2&0x70)
                                        {
                                                case 0x20: //midrow attribute change
                                                        switch (b2&0x0e)
                                                        {
                                                                case 0x00: //italics off
                                                                        strcat(ccbuf[ccmode],"\33[0m ");
                                                                        break;
                                                                case 0x0e: //italics on
                                                                        strcat(ccbuf[ccmode],"\33[36m ");
                                                                        break;
                                                        }
                                                        if (b2&0x01) //underline
                                                                strcat(ccbuf[ccmode],"\33[4m");
                                                        else
                                                                strcat(ccbuf[ccmode],"\33[24m");
                                                        break;
                                                case 0x30: //special character..
                                                        strcat(ccbuf[ccmode],specialchar[b2&0x0f]);
                                                        break;
                                        }
                                        break;
                                case 0x04:      //misc
                                case 0x05:      //misc + F
                                      //printf("ccmode %d cmd %02x\n",ccmode,b2);
                                        switch (b2)
                                        {
                                                case 0x21: //backspace
                                                        ccbuf[ccmode][len--]=0;
                                                        break;

                                                /* these codes are insignifigant if we're ignoring positioning */
                                                case 0x25: //2 row caption
                                                case 0x26: //3 row caption
                                                case 0x27: //4 row caption
                                                case 0x29: //resume direct caption
                                                case 0x2B: //resume text display
                                                case 0x2C: //erase displayed memory
                                                        break;

                                                case 0x2D: //carriage return
                                                        if (ccmode==2)
                                                                break;
                                                case 0x2F: //end caption + swap memory
                                                case 0x20: //resume caption (new caption)
                                                        if (!strlen(ccbuf[ccmode]))
                                                                        break;
                                                        for (x=0;x<strlen(ccbuf[ccmode]);x++)
                                                                for (y=0;y<keywords;y++)
                                                                        if (!strncasecmp(keyword[y], ccbuf[ccmode]+x, strlen(keyword[y])))
                                                        //printf("%s\n",ccbuf[ccmode]);
                                                        /* FALL */
                                                case 0x2A: //text restart
                                                case 0x2E: //erase non-displayed memory
                                                        memset(ccbuf[ccmode],0,255);
                                                        break;
                                        }
                                        break;
                                case 0x07:      //misc (TAB)
                                        for(x=0;x<(b2&0x03);x++)
                                                ccbuf[ccmode][len++]=' ';
                                        break;
                        }
                }
        }
        return 0;
}

int parityok(int n)     /* check parity for 2 bytes packed in n */
{
    int mask=0;
    int j, k;
    for (k = 1, j = 0; j < 7; j++)
        if (n & (1<<j)) k++;
    if ((k & 1) == ((n>>7)&1))
            mask|=0x00FF;
    for (k = 1, j = 8; j < 15; j++)
        if (n & (1<<j)) k++;
    if ((k & 1) == ((n>>15)&1))
            mask|=0xFF00;
   return mask;
}

int decodebit(unsigned char *data, int threshold)
{
    int i, sum = 0;
    for (i = 0; i < 23; i++)
        sum += data[i];
    return (sum > threshold*23);
}

int decode(unsigned char *vbiline)
{
    int max[7], min[7], val[7], i, clk, tmp, sample, packedbits = 0;

    for (clk=0; clk<7; clk++)
        max[clk] = min[clk] = val[clk] = -1;
    clk = tmp = 0;
    i=30;

    while (i < 600 && clk < 7) {        /* find and lock all 7 clocks */
        sample = vbiline[i];
        if (max[clk] < 0) { /* find maximum value before drop */
            if (sample > 85 && sample > val[clk])
                (val[clk] = sample, tmp = i);   /* mark new maximum found */
            else if (val[clk] - sample > 30)    /* far enough */
                (max[clk] = tmp, i = tmp + 10);
        } else { /* find minimum value after drop */
            if (sample < 85 && sample < val[clk])
                (val[clk] = sample, tmp = i);   /* mark new minimum found */
            else if (sample - val[clk] > 30)    /* searched far enough */
                (min[clk++] = tmp, i = tmp + 10);
        }
        i++;
    }

   i=min[6]=min[5]-max[5]+max[6];

    if (clk != 7 || vbiline[max[3]] - vbiline[min[5]] < 45)             /* failure to locate clock lead-in */
        return -1;

#ifdef XWIN
   for (clk=0;clk<7;clk++)
   {
           XDrawLine(dpy,Win,WinGC,min[clk]/2,0,min[clk]/2,128);
           XDrawLine(dpy,Win,WinGC1,max[clk]/2,0,max[clk]/2,128);
   }
   XFlush(dpy);
#endif


    /* calculate threshold */
    for (i=0,sample=0;i<7;i++)
            sample=(sample + vbiline[min[i]] + vbiline[max[i]])/3;

    for(i=min[6];vbiline[i]<sample;i++);

#ifdef XWIN
   for (clk=i;clk<i+57*18;clk+=57)
           XDrawLine(dpy,Win,WinGC,clk/2,0,clk/2,128);
   XFlush(dpy);
#endif


    tmp = i+57;
    for (i = 0; i < 16; i++)
        if(decodebit(&vbiline[tmp + i * 57], sample))
            packedbits |= 1<<i;
    return packedbits&parityok(packedbits);
} /* decode */


static void decode_cc(struct ivtv_sliced_data *s)
{
	static int xds_transport = 0;
	char c = s->data[0] & 0x7F;
	static char cc[CC_SIZE + 1];
	static int cc_idx;

	if (s->line >= 270) { /* field 2 */
		/* 0x01xx..0x0Exx ASCII_or_NUL[0..32] 0x0Fchks */
		if (odd_parity(s->data[0]) && (c >= 0x01 && c <= 0x0F)) {
			decode_xds(s);
			xds_transport = (c != 0x0F);
		} else if (xds_transport) {
			decode_xds(s);
		}
		return;
	}

	c = odd_parity(s->data[0]) ? s->data[0] & 0x7F : '?';
	c = printable(c);
	if (cc_idx >= CC_SIZE) {
		cc_idx = CC_SIZE - 2;
		memcpy(cc, cc + 2, cc_idx);
	} 
	cc[cc_idx++] = c;
	c = odd_parity(s->data[1]) ? s->data[1] & 0x7F : '?';
	c = printable(c);
	cc[cc_idx++] = c;
	cc[cc_idx] = 0;

	CCdecode(s->data[0], s->data[1], cc_idx);

	if (cc_idx == CC_SIZE) {
		printf("CC: %02x %02x: %s\n", s->data[0], s->data[1], cc);
		memset(cc, 0, CC_SIZE);
		cc_idx = 0;

	} 
}

const uint8_t
vbi_bit_reverse[256] = {
	0x00, 0x80, 0x40, 0xc0, 0x20, 0xa0, 0x60, 0xe0,	
	0x10, 0x90, 0x50, 0xd0, 0x30, 0xb0, 0x70, 0xf0,
	0x08, 0x88, 0x48, 0xc8, 0x28, 0xa8, 0x68, 0xe8,	
	0x18, 0x98, 0x58, 0xd8, 0x38, 0xb8, 0x78, 0xf8,
	0x04, 0x84, 0x44, 0xc4, 0x24, 0xa4, 0x64, 0xe4,	
	0x14, 0x94, 0x54, 0xd4, 0x34, 0xb4, 0x74, 0xf4,
	0x0c, 0x8c, 0x4c, 0xcc, 0x2c, 0xac, 0x6c, 0xec,	
	0x1c, 0x9c, 0x5c, 0xdc, 0x3c, 0xbc, 0x7c, 0xfc,
	0x02, 0x82, 0x42, 0xc2, 0x22, 0xa2, 0x62, 0xe2,	
	0x12, 0x92, 0x52, 0xd2, 0x32, 0xb2, 0x72, 0xf2,
	0x0a, 0x8a, 0x4a, 0xca, 0x2a, 0xaa, 0x6a, 0xea,	
	0x1a, 0x9a, 0x5a, 0xda, 0x3a, 0xba, 0x7a, 0xfa,
	0x06, 0x86, 0x46, 0xc6, 0x26, 0xa6, 0x66, 0xe6,	
	0x16, 0x96, 0x56, 0xd6, 0x36, 0xb6, 0x76, 0xf6,
	0x0e, 0x8e, 0x4e, 0xce, 0x2e, 0xae, 0x6e, 0xee,	
	0x1e, 0x9e, 0x5e, 0xde, 0x3e, 0xbe, 0x7e, 0xfe,
	0x01, 0x81, 0x41, 0xc1, 0x21, 0xa1, 0x61, 0xe1,	
	0x11, 0x91, 0x51, 0xd1, 0x31, 0xb1, 0x71, 0xf1,
	0x09, 0x89, 0x49, 0xc9, 0x29, 0xa9, 0x69, 0xe9,	
	0x19, 0x99, 0x59, 0xd9, 0x39, 0xb9, 0x79, 0xf9,
	0x05, 0x85, 0x45, 0xc5, 0x25, 0xa5, 0x65, 0xe5,	
	0x15, 0x95, 0x55, 0xd5, 0x35, 0xb5, 0x75, 0xf5,
	0x0d, 0x8d, 0x4d, 0xcd, 0x2d, 0xad, 0x6d, 0xed,	
	0x1d, 0x9d, 0x5d, 0xdd, 0x3d, 0xbd, 0x7d, 0xfd,
	0x03, 0x83, 0x43, 0xc3, 0x23, 0xa3, 0x63, 0xe3,	
	0x13, 0x93, 0x53, 0xd3, 0x33, 0xb3, 0x73, 0xf3,
	0x0b, 0x8b, 0x4b, 0xcb, 0x2b, 0xab, 0x6b, 0xeb,	
	0x1b, 0x9b, 0x5b, 0xdb, 0x3b, 0xbb, 0x7b, 0xfb,
	0x07, 0x87, 0x47, 0xc7, 0x27, 0xa7, 0x67, 0xe7,	
	0x17, 0x97, 0x57, 0xd7, 0x37, 0xb7, 0x77, 0xf7,
	0x0f, 0x8f, 0x4f, 0xcf, 0x2f, 0xaf, 0x6f, 0xef,	
	0x1f, 0x9f, 0x5f, 0xdf, 0x3f, 0xbf, 0x7f, 0xff
};

#define printable(c) ((((c) & 0x7F) < 0x20 || ((c) & 0x7F) > 0x7E) ? '.' : ((c) & 0x7F))

#define PIL(day, mon, hour, min) \
	(((day) << 15) + ((mon) << 11) + ((hour) << 6) + ((min) << 0))

static void dump_pil(int pil)
{
	int day, mon, hour, min;

	day = pil >> 15;
	mon = (pil >> 11) & 0xF;
	hour = (pil >> 6) & 0x1F;
	min = pil & 0x3F;

	if (pil == PIL(0, 15, 31, 63))
		printf(" PDC: Timer-control (no PDC)\n");
	else if (pil == PIL(0, 15, 30, 63))
		printf(" PDC: Recording inhibit/terminate\n");
	else if (pil == PIL(0, 15, 29, 63))
		printf(" PDC: Interruption\n");
	else if (pil == PIL(0, 15, 28, 63))
		printf(" PDC: Continue\n");
	else if (pil == PIL(31, 15, 31, 63))
		printf(" PDC: No time\n");
	else
		printf(" PDC: %05x, 200X-%02d-%02d %02d:%02d\n",
			pil, mon, day, hour, min);
}

void decode_vps(struct ivtv_sliced_data *s)
{
	static char pr_label[20];
	static char label[20];
	static int l = 0;
	int cni, pcs, pty, pil;
	int c;
	unsigned char *buf = s->data;

	printf("\nVPS:\n");

	c = vbi_bit_reverse[buf[1]];

	if ((int8_t) c < 0) {
		label[l] = 0;
		memcpy(pr_label, label, sizeof(pr_label));
		l = 0;
	}

	c &= 0x7F;

	label[l] = printable(c);

	l = (l + 1) % 16;

	printf(" 3-10: %02x %02x %02x %02x %02x %02x %02x %02x (\"%s\")\n",
		buf[0], buf[1], buf[2], buf[3], buf[4], buf[5], buf[6], buf[7], pr_label);

	pcs = buf[2] >> 6;

	cni = + ((buf[10] & 3) << 10)
	      + ((buf[11] & 0xC0) << 2)
	      + ((buf[8] & 0xC0) << 0)
	      + (buf[11] & 0x3F);

	pil = ((buf[8] & 0x3F) << 14) + (buf[9] << 6) + (buf[10] >> 2);

	pty = buf[12];

	printf(" CNI: %04x PCS: %d PTY: %d ", cni, pcs, pty);

	dump_pil(pil);
}

void process(struct ivtv_sliced_data *s)
{
	int i;

	if (s->line == 0) return;
	for (i = 0; i < 4; i++) if (s->data[i] != 0xa0) break;
	if (i == 4) return;
//	printf("line %03lu type %x\n", s->line, s->id);
	switch (s->id) {
		case IVTV_SLICED_TELETEXT_B:
			printf("teletext\n");
			break;
		case IVTV_SLICED_VPS:
			printf("vps\n");
			decode_vps(s);
			break;
		case IVTV_SLICED_WSS_625:
			printf("wss\n");
			decode_wss(s);
			break;
		case IVTV_SLICED_CAPTION_525:
		case IVTV_SLICED_CAPTION_625:
//			printf("cc\n");
			decode_cc(s);
			break;
		default:
			printf("unknown\n");
			break;
	}
}

int main(int argc, char **argv)
{
	char *device = "/dev/vbi0";
	struct ivtv_sliced_vbi_format fmt;
	char *buf;
	int fh;

	if (argc == 2) device = argv[1];
	fh = open(device, O_RDONLY);

	if (fh == -1) {
		fprintf(stderr, "cannot open %s\n", device);
		return 1;
	}

	fmt.service_set = ~0;
	if (ioctl(fh, IVTV_IOC_S_VBI_MODE, &fmt) < 0) {
		perror("vbi");
		close(fh);
		return 1;
	}

	printf("%08lx, %ld, %ld\n", fmt.service_set, fmt.packet_size, fmt.io_size);
	buf = malloc(fmt.io_size);
	for (;;) {
		int size = read(fh, buf, fmt.io_size);
		int i;
	
	 	if (size <= 0) break;	
		for (i = 0; i < size / fmt.packet_size; i++) {
			process((struct ivtv_sliced_data *)(buf + i * fmt.packet_size));
		}
	}
	close(fh);
	return 0;
}

